"""
benchmarks is a standard framework for benchmarking the performance of models,
and for methods development
"""

from typing import *

import pandas as pd
from tqdm import tqdm

import olorenchemengine as oce
from olorenchemengine import *
from olorenchemengine.base_class import *
from olorenchemengine.manager import *
from olorenchemengine.visualizations import *


class BenchmarkDatasets(BaseVisualization):
    """
    Track and visualize the performance of models on user-defined datasets.

    Parameters:
        datasets List[BaseDataset] (optional): List of dataset in BaseDataset format to benchmark against.

    Methods:
        run(models: Union[BaseModel, List[BaseModel]]): Benchmark the given model or list of models against the given dataset(s).
        render, render_ipynb() (Inherited BaseVisualization methods): Render a plot of the benchmark results, showing the number of datasets each model is superior on.

    Example
    ------------------------------
    import olorenchemengine as oce

    models = [
        oce.RandomForestModel(representation=oce.MorganVecRepresentation(radius=2, nbits=2048)),
        oce.XGBoostModel(oce.MorganVecRepresentation(radius=2, nbits=2048))
    ]

    benchmark = oce.BenchmarkDatasets(datasets=['''Your Datasets Here'''])
    benchmark.run(models)
    benchmark.render_ipynb()
    ------------------------------
    """

    @log_arguments
    def __init__(self, datasets: List[BaseDataset], log=True):
        super().__init__(log=False)
        self.packages = ["plotly"]

        self.datasets = datasets
        self.metrics = [
            "ROC-AUC"
            if detect_setting(dataset.train_dataset[1]) == "classification"
            else "Root Mean Squared Error"
            for dataset in datasets
        ]
        self.managers = [
            ModelManager(dataset, metrics=[metric])
            for dataset, metric in zip(self.datasets, self.metrics)
        ]
        self.model_names = []
        self.model_params = []

    def run(self, models: Union[BaseModel, List[BaseModel]]):
        if not isinstance(models, list):
            models = [models]

        for model in models:
            print("Running benchmark on model: {}".format(model.name))
            for manager in tqdm(self.managers):
                manager.run(model)
            self.model_names.append(model.name)
            self.model_params.append(oce.pretty_params_str(model))

    @property
    def JS_NAME(self):
        return "BenchmarkDatasets"

    def get_data(self):
        d = {model_name: 0 for model_name in self.model_names}

        benchmark_data = pd.DataFrame()
        for metric, manager in zip(self.metrics, self.managers):
            direction = metric_direction[metric]

            model_database = manager.get_model_database()
            model_database["Dataset Name"] = manager.dataset.name
            benchmark_data = benchmark_data.append(model_database)

            if direction == "higher":
                best_row = model_database[
                    model_database[metric] == model_database[metric].max()
                ].iloc[0]
            elif direction == "lower":
                best_row = model_database[
                    model_database[metric] == model_database[metric].min()
                ].iloc[0]

            d[best_row["Model Name"]] = d[best_row["Model Name"]] + 1

        out = dict()
        out["model_names"] = self.model_names
        out["model_params"] = [
            str(mp).replace("\n", "<br>")
            for mn, mp in zip(self.model_names, self.model_params)
        ]
        out["model_counts"] = [d[name] for name in self.model_names]
        return out

    def _save(self):
        d = super()._save()
        d["managers"] = [manager._save() for manager in self.managers]
        d["model_names"] = self.model_names
        d["model_params"] = self.model_params
        return d

    def _load(self, d):
        super()._load(d)
        for manager, manager_d in zip(self.managers, d["managers"]):
            manager._load(manager_d)
        self.model_names = d["model_names"]
        self.model_params = d["model_params"]


class BenchmarkTDC(BenchmarkDatasets):
    """
    Track and visualize the performance of models on selected ADME datasets from the Therapeutics Data Commons (TDC). https://tdcommons.ai/

    Parameters:
        datasets List[str] (optional): List of TDC dataset names to benchmark. If None, all datasets are benchmarked. Dataset names are
        listed in self.metrics_dict and @ https://tdcommons.ai/benchmark/admet_group/01caco2/.

    Methods:
        run(models: Union[BaseModel, List[BaseModel]]): Benchmark the given model or list of models against the specified dataset(s).
        get_performance(): Returns a pandas dataframe with the performance metrics of the models on the selected datasets. Can only be called after benchmark is run with "run" method.
        render, render_ipynb() (Inherited BaseVisualization methods): Render a plot of the benchmark results, showing the number of datasets each model is superior on.

    Example
    ------------------------------
    import olorenchemengine as oce

    models = [
        oce.RandomForestModel(representation=oce.MorganVecRepresentation(radius=2, nbits=2048)),
        oce.XGBoostModel(oce.MorganVecRepresentation(radius=2, nbits=2048))
    ]

    benchmark_tdc = oce.BenchmarkTDC(datasets=['HIA_Hou', 'Pgp_Broccatelli', 'BBB_Martins'])
    benchmark_tdc.run(models)
    benchmark_tdc.render_ipynb()
    ------------------------------
    """

    @log_arguments
    def __init__(self, datasets: List[str] = ["all"], log=True):
        self.packages = ["plotly"]
        from tdc.benchmark_group import admet_group

        self.metrics_dict = {
            "HIA_Hou": "ROC-AUC",
            "Pgp_Broccatelli": "ROC-AUC",
            "Bioavailability_Ma": "ROC-AUC",
            "BBB_Martins": "ROC-AUC",
            "hERG": "ROC-AUC",
            "AMES": "ROC-AUC",
            "DILI": "ROC-AUC",
            "CYP3A4_Substrate_CarbonMangels": "ROC-AUC",
            "Lipophilicity_AstraZeneca": "Mean Absolute Error",
            "Solubility_AqSolDB": "Mean Absolute Error",
            "LD50_Zhu": "Mean Absolute Error",
            "Caco2_Wang": "Mean Absolute Error",
            "PPBR_AZ": "Mean Absolute Error",
            "CYP2C9_Veith": "Average Precision",
            "CYP2D6_Veith": "Average Precision",
            "CYP3A4_Veith": "Average Precision",
            "CYP2C9_Substrate_CarbonMangels": "Average Precision",
            "CYP2D6_Substrate_CarbonMangels": "Average Precision",
            "Half_Life_Obach": "Spearman",
            "Clearance_Hepatocyte_AZ": "Spearman",
            "Clearance_Microsome_AZ": "Spearman",
            "VDss_Lombardo": "Spearman",
        }

        if datasets == ["all"]:
            datasets = list(self.metrics_dict.keys())

        group = admet_group(path="data/")
        datasets_wrapped = []
        self.metrics = []
        for name in datasets:
            benchmark = group.get(name)
            test = benchmark["test"]
            train, valid = group.get_train_valid_split(
                benchmark=name, split_type="default", seed=0
            )
            train["split"] = "train"
            test["split"] = "test"
            self.metrics.append(self.metrics_dict[name])

            data = pd.concat([train, test]).to_csv()
            datasets_wrapped.append(
                BaseDataset(data=data, structure_col="Drug", property_col="Y")
            )
            i = datasets.index(name)
            datasets_wrapped[i].train = train
            datasets_wrapped[i].test = test
            datasets_wrapped[i].data = pd.concat(
                [datasets_wrapped[i].train, datasets_wrapped[i].test]
            )

        self.dataset_names = datasets
        self.datasets = datasets_wrapped
        self.managers = [
            ModelManager(dataset, metrics=[metric])
            for dataset, metric in zip(self.datasets, self.metrics)
        ]
        self.model_names = []
        self.model_params = []

    """Returns DataFrame of perfomance metric for each dataset and model benchmarked."""

    def get_performance(self):
        out = pd.DataFrame(data=None, columns=["Dataset", "Metric"])
        out["Dataset"] = self.dataset_names
        out["Metric"] = [self.metrics_dict[name] for name in self.dataset_names]
        for name in self.model_names:
            out[name] = np.nan
        for manager in self.managers:
            dataset_name = self.dataset_names[self.managers.index(manager)]
            manager_data = manager.model_database
            for i, row in manager_data.iterrows():
                index = i
                out.loc[out["Dataset"] == dataset_name, row["Model Name"]] = row[
                    self.metrics_dict[dataset_name]
                ]
        return out


class BenchmarkMolNet(BenchmarkDatasets):
    """
    Track and visualize the performance of models on selected datasets from the MoleculeNet collection. https://moleculenet.org/

    Parameters:
        datasets List[str] (optional): List of MoleculeNet dataset names to benchmark. If None, all datasets are benchmarked. Dataset names are
            listed in self.datasets_loaded of the init function.
        filepath (str): filepath to save the benchmark results
        mode (str): "default" is those given by MoleculeNet direectly, "geognn"
            is those used by GeoGNN (https://www.nature.com/articles/s42256-021-00438-4) and
            append the string "-small" to use a subset of 5 tasks for benchmarks with more
            than five tasks.
        eval (str): whether to output model performance against the validation set ("valid")
            or the test set ("test").

    Methods:
        run(models: Union[BaseModel, List[BaseModel]]): Benchmark the given model or list of models against the specified dataset(s).
        render, render_ipynb() (Inherited BaseVisualization methods): Render a plot of the benchmark results, showing the number of datasets each model is superior on.

    Example
    ------------------------------
    import olorenchemengine as oce

    models = [
        oce.RandomForestModel(representation=oce.MorganVecRepresentation(radius=2, nbits=2048)),
        oce.XGBoostModel(oce.MorganVecRepresentation(radius=2, nbits=2048))
    ]

    benchmark_mnet = oce.BenchmarkMolNet(datasets=['bace_classification', 'clintox', 'sider'])
    benchmark_mnet.run(models)
    benchmark_mnet.render_ipynb()
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        datasets: List[str] = ["all"],
        file_path=None,
        mode="default",
        eval="valid",
        log=True,
    ):
        self.mode = mode
        self.eval = eval
        if "default" in self.mode or "small" in self.mode:
            self.datasets_loaded = {
                "bace_classification": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_bace_classification.csv",
                "bbbp": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_bbbp.csv",
                "clintox": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_clintox.csv",
                "hiv": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_hiv.csv",
                "muv": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_muv.csv",
                "pcba": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_pcba.csv",
                "sider": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_sider.csv",
                "tox21": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_tox21.csv",
                "toxcast": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_toxcast.csv",
                "delaney_esol": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_delaney.csv",
                "freesolv": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_freesolv.csv",
                "lipo": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_lipo.csv",
                "bace_regression": "https://storage.googleapis.com/oloren-public-data/MoleculeNet/load_bace_regression.csv",
            }
        elif "geognn" in self.mode:
            self.datasets_loaded = {
                "bace": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/bace.csv",
                "bbbp": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/BBBP.csv",
                "clintox": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/clintox.csv",
                "sider": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/sider.csv",
                "tox21": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/tox21.csv",
                "toxcast": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/toxcast_data.csv",
                "delaney_esol": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/delaney-processed.csv",
                "freesolv": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/freesolv.csv",
                "lipo": "https://storage.googleapis.com/oloren-public-data/MoleculeNet2/Lipophilicity.csv",
            }
        else:
            assert (
                False
            ), "Invalid mode, should be either 'default', 'small' or 'geognn' with optional attribute small"

        self.datasets_types = {
            "bace_classification": "classification",
            "bace": "classification",
            "bbbp": "classification",
            "clintox": "classification",
            "hiv": "classification",
            "muv": "classification",
            "pcba": "classification",
            "sider": "classification",
            "tox21": "classification",
            "toxcast": "classification",
            "delaney_esol": "regression",
            "freesolv": "regression",
            "lipo": "regression",
            "bace_regression": "regression",
        }

        self.dataset_property_cols = {
            "bace_classification": "Class",
            "bace": "Class",
            "bbbp": "p_np",
            "clintox": ["FDA_APPROVED", "CT_TOX"],
            "hiv": "HIV_active",
            "muv": [
                "MUV-466",
                "MUV-548",
                "MUV-600",
                "MUV-644",
                "MUV-652",
                "MUV-689",
                "MUV-692",
                "MUV-712",
                "MUV-713",
                "MUV-733",
                "MUV-737",
                "MUV-810",
                "MUV-832",
                "MUV-846",
                "MUV-852",
                "MUV-858",
                "MUV-859",
            ],
            "pcba": [
                "PCBA-1030",
                "PCBA-1379",
                "PCBA-1452",
                "PCBA-1454",
                "PCBA-1457",
                "PCBA-1458",
                "PCBA-1460",
                "PCBA-1461",
                "PCBA-1468",
                "PCBA-1469",
                "PCBA-1471",
                "PCBA-1479",
                "PCBA-1631",
                "PCBA-1634",
                "PCBA-1688",
                "PCBA-1721",
                "PCBA-2100",
                "PCBA-2101",
                "PCBA-2147",
                "PCBA-2242",
                "PCBA-2326",
                "PCBA-2451",
                "PCBA-2517",
                "PCBA-2528",
                "PCBA-2546",
                "PCBA-2549",
                "PCBA-2551",
                "PCBA-2662",
                "PCBA-2675",
                "PCBA-2676",
                "PCBA-411",
                "PCBA-463254",
                "PCBA-485281",
                "PCBA-485290",
                "PCBA-485294",
                "PCBA-485297",
                "PCBA-485313",
                "PCBA-485314",
                "PCBA-485341",
                "PCBA-485349",
                "PCBA-485353",
                "PCBA-485360",
                "PCBA-485364",
                "PCBA-485367",
                "PCBA-492947",
                "PCBA-493208",
                "PCBA-504327",
                "PCBA-504332",
                "PCBA-504333",
                "PCBA-504339",
                "PCBA-504444",
                "PCBA-504466",
                "PCBA-504467",
                "PCBA-504706",
                "PCBA-504842",
                "PCBA-504845",
                "PCBA-504847",
                "PCBA-504891",
                "PCBA-540276",
                "PCBA-540317",
                "PCBA-588342",
                "PCBA-588453",
                "PCBA-588456",
                "PCBA-588579",
                "PCBA-588590",
                "PCBA-588591",
                "PCBA-588795",
                "PCBA-588855",
                "PCBA-602179",
                "PCBA-602233",
                "PCBA-602310",
                "PCBA-602313",
                "PCBA-602332",
                "PCBA-624170",
                "PCBA-624171",
                "PCBA-624173",
                "PCBA-624202",
                "PCBA-624246",
                "PCBA-624287",
                "PCBA-624288",
                "PCBA-624291",
                "PCBA-624296",
                "PCBA-624297",
                "PCBA-624417",
                "PCBA-651635",
                "PCBA-651644",
                "PCBA-651768",
                "PCBA-651965",
                "PCBA-652025",
                "PCBA-652104",
                "PCBA-652105",
                "PCBA-652106",
                "PCBA-686970",
                "PCBA-686978",
                "PCBA-686979",
                "PCBA-720504",
                "PCBA-720532",
                "PCBA-720542",
                "PCBA-720551",
                "PCBA-720553",
                "PCBA-720579",
                "PCBA-720580",
                "PCBA-720707",
                "PCBA-720708",
                "PCBA-720709",
                "PCBA-720711",
                "PCBA-743255",
                "PCBA-743266",
                "PCBA-875",
                "PCBA-881",
                "PCBA-883",
                "PCBA-884",
                "PCBA-885",
                "PCBA-887",
                "PCBA-891",
                "PCBA-899",
                "PCBA-902",
                "PCBA-903",
                "PCBA-904",
                "PCBA-912",
                "PCBA-914",
                "PCBA-915",
                "PCBA-924",
                "PCBA-925",
                "PCBA-926",
                "PCBA-927",
                "PCBA-938",
                "PCBA-995",
            ],
            "sider": [
                "Hepatobiliary disorders",
                "Metabolism and nutrition disorders",
                "Product issues",
                "Eye disorders",
                "Investigations",
                "Musculoskeletal and connective tissue disorders",
                "Gastrointestinal disorders",
                "Social circumstances",
                "Immune system disorders",
                "Reproductive system and breast disorders",
                "Neoplasms benign, malignant and unspecified (incl cysts and polyps)",
                "General disorders and administration site conditions",
                "Endocrine disorders",
                "Surgical and medical procedures",
                "Vascular disorders",
                "Blood and lymphatic system disorders",
                "Skin and subcutaneous tissue disorders",
                "Congenital, familial and genetic disorders",
                "Infections and infestations",
                "Respiratory, thoracic and mediastinal disorders",
                "Psychiatric disorders",
                "Renal and urinary disorders",
                "Pregnancy, puerperium and perinatal conditions",
                "Ear and labyrinth disorders",
                "Cardiac disorders",
                "Nervous system disorders",
                "Injury, poisoning and procedural complications",
            ],
            "tox21": [
                "NR-AR",
                "NR-AR-LBD",
                "NR-AhR",
                "NR-Aromatase",
                "NR-ER",
                "NR-ER-LBD",
                "NR-PPAR-gamma",
                "SR-ARE",
                "SR-ATAD5",
                "SR-HSE",
                "SR-MMP",
                "SR-p53",
            ],
            "toxcast": [
                "ACEA_T47D_80hr_Negative",
                "ACEA_T47D_80hr_Positive",
                "APR_HepG2_CellCycleArrest_24h_dn",
                "APR_HepG2_CellCycleArrest_24h_up",
                "APR_HepG2_CellCycleArrest_72h_dn",
                "APR_HepG2_CellLoss_24h_dn",
                "APR_HepG2_CellLoss_72h_dn",
                "APR_HepG2_MicrotubuleCSK_24h_dn",
                "APR_HepG2_MicrotubuleCSK_24h_up",
                "APR_HepG2_MicrotubuleCSK_72h_dn",
                "APR_HepG2_MicrotubuleCSK_72h_up",
                "APR_HepG2_MitoMass_24h_dn",
                "APR_HepG2_MitoMass_24h_up",
                "APR_HepG2_MitoMass_72h_dn",
                "APR_HepG2_MitoMass_72h_up",
                "APR_HepG2_MitoMembPot_1h_dn",
                "APR_HepG2_MitoMembPot_24h_dn",
                "APR_HepG2_MitoMembPot_72h_dn",
                "APR_HepG2_MitoticArrest_24h_up",
                "APR_HepG2_MitoticArrest_72h_up",
                "APR_HepG2_NuclearSize_24h_dn",
                "APR_HepG2_NuclearSize_72h_dn",
                "APR_HepG2_NuclearSize_72h_up",
                "APR_HepG2_OxidativeStress_24h_up",
                "APR_HepG2_OxidativeStress_72h_up",
                "APR_HepG2_StressKinase_1h_up",
                "APR_HepG2_StressKinase_24h_up",
                "APR_HepG2_StressKinase_72h_up",
                "APR_HepG2_p53Act_24h_up",
                "APR_HepG2_p53Act_72h_up",
                "APR_Hepat_Apoptosis_24hr_up",
                "APR_Hepat_Apoptosis_48hr_up",
                "APR_Hepat_CellLoss_24hr_dn",
                "APR_Hepat_CellLoss_48hr_dn",
                "APR_Hepat_DNADamage_24hr_up",
                "APR_Hepat_DNADamage_48hr_up",
                "APR_Hepat_DNATexture_24hr_up",
                "APR_Hepat_DNATexture_48hr_up",
                "APR_Hepat_MitoFxnI_1hr_dn",
                "APR_Hepat_MitoFxnI_24hr_dn",
                "APR_Hepat_MitoFxnI_48hr_dn",
                "APR_Hepat_NuclearSize_24hr_dn",
                "APR_Hepat_NuclearSize_48hr_dn",
                "APR_Hepat_Steatosis_24hr_up",
                "APR_Hepat_Steatosis_48hr_up",
                "ATG_AP_1_CIS_dn",
                "ATG_AP_1_CIS_up",
                "ATG_AP_2_CIS_dn",
                "ATG_AP_2_CIS_up",
                "ATG_AR_TRANS_dn",
                "ATG_AR_TRANS_up",
                "ATG_Ahr_CIS_dn",
                "ATG_Ahr_CIS_up",
                "ATG_BRE_CIS_dn",
                "ATG_BRE_CIS_up",
                "ATG_CAR_TRANS_dn",
                "ATG_CAR_TRANS_up",
                "ATG_CMV_CIS_dn",
                "ATG_CMV_CIS_up",
                "ATG_CRE_CIS_dn",
                "ATG_CRE_CIS_up",
                "ATG_C_EBP_CIS_dn",
                "ATG_C_EBP_CIS_up",
                "ATG_DR4_LXR_CIS_dn",
                "ATG_DR4_LXR_CIS_up",
                "ATG_DR5_CIS_dn",
                "ATG_DR5_CIS_up",
                "ATG_E2F_CIS_dn",
                "ATG_E2F_CIS_up",
                "ATG_EGR_CIS_up",
                "ATG_ERE_CIS_dn",
                "ATG_ERE_CIS_up",
                "ATG_ERRa_TRANS_dn",
                "ATG_ERRg_TRANS_dn",
                "ATG_ERRg_TRANS_up",
                "ATG_ERa_TRANS_up",
                "ATG_E_Box_CIS_dn",
                "ATG_E_Box_CIS_up",
                "ATG_Ets_CIS_dn",
                "ATG_Ets_CIS_up",
                "ATG_FXR_TRANS_up",
                "ATG_FoxA2_CIS_dn",
                "ATG_FoxA2_CIS_up",
                "ATG_FoxO_CIS_dn",
                "ATG_FoxO_CIS_up",
                "ATG_GAL4_TRANS_dn",
                "ATG_GATA_CIS_dn",
                "ATG_GATA_CIS_up",
                "ATG_GLI_CIS_dn",
                "ATG_GLI_CIS_up",
                "ATG_GRE_CIS_dn",
                "ATG_GRE_CIS_up",
                "ATG_GR_TRANS_dn",
                "ATG_GR_TRANS_up",
                "ATG_HIF1a_CIS_dn",
                "ATG_HIF1a_CIS_up",
                "ATG_HNF4a_TRANS_dn",
                "ATG_HNF4a_TRANS_up",
                "ATG_HNF6_CIS_dn",
                "ATG_HNF6_CIS_up",
                "ATG_HSE_CIS_dn",
                "ATG_HSE_CIS_up",
                "ATG_IR1_CIS_dn",
                "ATG_IR1_CIS_up",
                "ATG_ISRE_CIS_dn",
                "ATG_ISRE_CIS_up",
                "ATG_LXRa_TRANS_dn",
                "ATG_LXRa_TRANS_up",
                "ATG_LXRb_TRANS_dn",
                "ATG_LXRb_TRANS_up",
                "ATG_MRE_CIS_up",
                "ATG_M_06_TRANS_up",
                "ATG_M_19_CIS_dn",
                "ATG_M_19_TRANS_dn",
                "ATG_M_19_TRANS_up",
                "ATG_M_32_CIS_dn",
                "ATG_M_32_CIS_up",
                "ATG_M_32_TRANS_dn",
                "ATG_M_32_TRANS_up",
                "ATG_M_61_TRANS_up",
                "ATG_Myb_CIS_dn",
                "ATG_Myb_CIS_up",
                "ATG_Myc_CIS_dn",
                "ATG_Myc_CIS_up",
                "ATG_NFI_CIS_dn",
                "ATG_NFI_CIS_up",
                "ATG_NF_kB_CIS_dn",
                "ATG_NF_kB_CIS_up",
                "ATG_NRF1_CIS_dn",
                "ATG_NRF1_CIS_up",
                "ATG_NRF2_ARE_CIS_dn",
                "ATG_NRF2_ARE_CIS_up",
                "ATG_NURR1_TRANS_dn",
                "ATG_NURR1_TRANS_up",
                "ATG_Oct_MLP_CIS_dn",
                "ATG_Oct_MLP_CIS_up",
                "ATG_PBREM_CIS_dn",
                "ATG_PBREM_CIS_up",
                "ATG_PPARa_TRANS_dn",
                "ATG_PPARa_TRANS_up",
                "ATG_PPARd_TRANS_up",
                "ATG_PPARg_TRANS_up",
                "ATG_PPRE_CIS_dn",
                "ATG_PPRE_CIS_up",
                "ATG_PXRE_CIS_dn",
                "ATG_PXRE_CIS_up",
                "ATG_PXR_TRANS_dn",
                "ATG_PXR_TRANS_up",
                "ATG_Pax6_CIS_up",
                "ATG_RARa_TRANS_dn",
                "ATG_RARa_TRANS_up",
                "ATG_RARb_TRANS_dn",
                "ATG_RARb_TRANS_up",
                "ATG_RARg_TRANS_dn",
                "ATG_RARg_TRANS_up",
                "ATG_RORE_CIS_dn",
                "ATG_RORE_CIS_up",
                "ATG_RORb_TRANS_dn",
                "ATG_RORg_TRANS_dn",
                "ATG_RORg_TRANS_up",
                "ATG_RXRa_TRANS_dn",
                "ATG_RXRa_TRANS_up",
                "ATG_RXRb_TRANS_dn",
                "ATG_RXRb_TRANS_up",
                "ATG_SREBP_CIS_dn",
                "ATG_SREBP_CIS_up",
                "ATG_STAT3_CIS_dn",
                "ATG_STAT3_CIS_up",
                "ATG_Sox_CIS_dn",
                "ATG_Sox_CIS_up",
                "ATG_Sp1_CIS_dn",
                "ATG_Sp1_CIS_up",
                "ATG_TAL_CIS_dn",
                "ATG_TAL_CIS_up",
                "ATG_TA_CIS_dn",
                "ATG_TA_CIS_up",
                "ATG_TCF_b_cat_CIS_dn",
                "ATG_TCF_b_cat_CIS_up",
                "ATG_TGFb_CIS_dn",
                "ATG_TGFb_CIS_up",
                "ATG_THRa1_TRANS_dn",
                "ATG_THRa1_TRANS_up",
                "ATG_VDRE_CIS_dn",
                "ATG_VDRE_CIS_up",
                "ATG_VDR_TRANS_dn",
                "ATG_VDR_TRANS_up",
                "ATG_XTT_Cytotoxicity_up",
                "ATG_Xbp1_CIS_dn",
                "ATG_Xbp1_CIS_up",
                "ATG_p53_CIS_dn",
                "ATG_p53_CIS_up",
                "BSK_3C_Eselectin_down",
                "BSK_3C_HLADR_down",
                "BSK_3C_ICAM1_down",
                "BSK_3C_IL8_down",
                "BSK_3C_MCP1_down",
                "BSK_3C_MIG_down",
                "BSK_3C_Proliferation_down",
                "BSK_3C_SRB_down",
                "BSK_3C_Thrombomodulin_down",
                "BSK_3C_Thrombomodulin_up",
                "BSK_3C_TissueFactor_down",
                "BSK_3C_TissueFactor_up",
                "BSK_3C_VCAM1_down",
                "BSK_3C_Vis_down",
                "BSK_3C_uPAR_down",
                "BSK_4H_Eotaxin3_down",
                "BSK_4H_MCP1_down",
                "BSK_4H_Pselectin_down",
                "BSK_4H_Pselectin_up",
                "BSK_4H_SRB_down",
                "BSK_4H_VCAM1_down",
                "BSK_4H_VEGFRII_down",
                "BSK_4H_uPAR_down",
                "BSK_4H_uPAR_up",
                "BSK_BE3C_HLADR_down",
                "BSK_BE3C_IL1a_down",
                "BSK_BE3C_IP10_down",
                "BSK_BE3C_MIG_down",
                "BSK_BE3C_MMP1_down",
                "BSK_BE3C_MMP1_up",
                "BSK_BE3C_PAI1_down",
                "BSK_BE3C_SRB_down",
                "BSK_BE3C_TGFb1_down",
                "BSK_BE3C_tPA_down",
                "BSK_BE3C_uPAR_down",
                "BSK_BE3C_uPAR_up",
                "BSK_BE3C_uPA_down",
                "BSK_CASM3C_HLADR_down",
                "BSK_CASM3C_IL6_down",
                "BSK_CASM3C_IL6_up",
                "BSK_CASM3C_IL8_down",
                "BSK_CASM3C_LDLR_down",
                "BSK_CASM3C_LDLR_up",
                "BSK_CASM3C_MCP1_down",
                "BSK_CASM3C_MCP1_up",
                "BSK_CASM3C_MCSF_down",
                "BSK_CASM3C_MCSF_up",
                "BSK_CASM3C_MIG_down",
                "BSK_CASM3C_Proliferation_down",
                "BSK_CASM3C_Proliferation_up",
                "BSK_CASM3C_SAA_down",
                "BSK_CASM3C_SAA_up",
                "BSK_CASM3C_SRB_down",
                "BSK_CASM3C_Thrombomodulin_down",
                "BSK_CASM3C_Thrombomodulin_up",
                "BSK_CASM3C_TissueFactor_down",
                "BSK_CASM3C_VCAM1_down",
                "BSK_CASM3C_VCAM1_up",
                "BSK_CASM3C_uPAR_down",
                "BSK_CASM3C_uPAR_up",
                "BSK_KF3CT_ICAM1_down",
                "BSK_KF3CT_IL1a_down",
                "BSK_KF3CT_IP10_down",
                "BSK_KF3CT_IP10_up",
                "BSK_KF3CT_MCP1_down",
                "BSK_KF3CT_MCP1_up",
                "BSK_KF3CT_MMP9_down",
                "BSK_KF3CT_SRB_down",
                "BSK_KF3CT_TGFb1_down",
                "BSK_KF3CT_TIMP2_down",
                "BSK_KF3CT_uPA_down",
                "BSK_LPS_CD40_down",
                "BSK_LPS_Eselectin_down",
                "BSK_LPS_Eselectin_up",
                "BSK_LPS_IL1a_down",
                "BSK_LPS_IL1a_up",
                "BSK_LPS_IL8_down",
                "BSK_LPS_IL8_up",
                "BSK_LPS_MCP1_down",
                "BSK_LPS_MCSF_down",
                "BSK_LPS_PGE2_down",
                "BSK_LPS_PGE2_up",
                "BSK_LPS_SRB_down",
                "BSK_LPS_TNFa_down",
                "BSK_LPS_TNFa_up",
                "BSK_LPS_TissueFactor_down",
                "BSK_LPS_TissueFactor_up",
                "BSK_LPS_VCAM1_down",
                "BSK_SAg_CD38_down",
                "BSK_SAg_CD40_down",
                "BSK_SAg_CD69_down",
                "BSK_SAg_Eselectin_down",
                "BSK_SAg_Eselectin_up",
                "BSK_SAg_IL8_down",
                "BSK_SAg_IL8_up",
                "BSK_SAg_MCP1_down",
                "BSK_SAg_MIG_down",
                "BSK_SAg_PBMCCytotoxicity_down",
                "BSK_SAg_PBMCCytotoxicity_up",
                "BSK_SAg_Proliferation_down",
                "BSK_SAg_SRB_down",
                "BSK_hDFCGF_CollagenIII_down",
                "BSK_hDFCGF_EGFR_down",
                "BSK_hDFCGF_EGFR_up",
                "BSK_hDFCGF_IL8_down",
                "BSK_hDFCGF_IP10_down",
                "BSK_hDFCGF_MCSF_down",
                "BSK_hDFCGF_MIG_down",
                "BSK_hDFCGF_MMP1_down",
                "BSK_hDFCGF_MMP1_up",
                "BSK_hDFCGF_PAI1_down",
                "BSK_hDFCGF_Proliferation_down",
                "BSK_hDFCGF_SRB_down",
                "BSK_hDFCGF_TIMP1_down",
                "BSK_hDFCGF_VCAM1_down",
                "CEETOX_H295R_11DCORT_dn",
                "CEETOX_H295R_ANDR_dn",
                "CEETOX_H295R_CORTISOL_dn",
                "CEETOX_H295R_DOC_dn",
                "CEETOX_H295R_DOC_up",
                "CEETOX_H295R_ESTRADIOL_dn",
                "CEETOX_H295R_ESTRADIOL_up",
                "CEETOX_H295R_ESTRONE_dn",
                "CEETOX_H295R_ESTRONE_up",
                "CEETOX_H295R_OHPREG_up",
                "CEETOX_H295R_OHPROG_dn",
                "CEETOX_H295R_OHPROG_up",
                "CEETOX_H295R_PROG_up",
                "CEETOX_H295R_TESTO_dn",
                "CLD_ABCB1_48hr",
                "CLD_ABCG2_48hr",
                "CLD_CYP1A1_24hr",
                "CLD_CYP1A1_48hr",
                "CLD_CYP1A1_6hr",
                "CLD_CYP1A2_24hr",
                "CLD_CYP1A2_48hr",
                "CLD_CYP1A2_6hr",
                "CLD_CYP2B6_24hr",
                "CLD_CYP2B6_48hr",
                "CLD_CYP2B6_6hr",
                "CLD_CYP3A4_24hr",
                "CLD_CYP3A4_48hr",
                "CLD_CYP3A4_6hr",
                "CLD_GSTA2_48hr",
                "CLD_SULT2A_24hr",
                "CLD_SULT2A_48hr",
                "CLD_UGT1A1_24hr",
                "CLD_UGT1A1_48hr",
                "NCCT_HEK293T_CellTiterGLO",
                "NCCT_QuantiLum_inhib_2_dn",
                "NCCT_QuantiLum_inhib_dn",
                "NCCT_TPO_AUR_dn",
                "NCCT_TPO_GUA_dn",
                "NHEERL_ZF_144hpf_TERATOSCORE_up",
                "NVS_ADME_hCYP19A1",
                "NVS_ADME_hCYP1A1",
                "NVS_ADME_hCYP1A2",
                "NVS_ADME_hCYP2A6",
                "NVS_ADME_hCYP2B6",
                "NVS_ADME_hCYP2C19",
                "NVS_ADME_hCYP2C9",
                "NVS_ADME_hCYP2D6",
                "NVS_ADME_hCYP3A4",
                "NVS_ADME_hCYP4F12",
                "NVS_ADME_rCYP2C12",
                "NVS_ENZ_hAChE",
                "NVS_ENZ_hAMPKa1",
                "NVS_ENZ_hAurA",
                "NVS_ENZ_hBACE",
                "NVS_ENZ_hCASP5",
                "NVS_ENZ_hCK1D",
                "NVS_ENZ_hDUSP3",
                "NVS_ENZ_hES",
                "NVS_ENZ_hElastase",
                "NVS_ENZ_hFGFR1",
                "NVS_ENZ_hGSK3b",
                "NVS_ENZ_hMMP1",
                "NVS_ENZ_hMMP13",
                "NVS_ENZ_hMMP2",
                "NVS_ENZ_hMMP3",
                "NVS_ENZ_hMMP7",
                "NVS_ENZ_hMMP9",
                "NVS_ENZ_hPDE10",
                "NVS_ENZ_hPDE4A1",
                "NVS_ENZ_hPDE5",
                "NVS_ENZ_hPI3Ka",
                "NVS_ENZ_hPTEN",
                "NVS_ENZ_hPTPN11",
                "NVS_ENZ_hPTPN12",
                "NVS_ENZ_hPTPN13",
                "NVS_ENZ_hPTPN9",
                "NVS_ENZ_hPTPRC",
                "NVS_ENZ_hSIRT1",
                "NVS_ENZ_hSIRT2",
                "NVS_ENZ_hTrkA",
                "NVS_ENZ_hVEGFR2",
                "NVS_ENZ_oCOX1",
                "NVS_ENZ_oCOX2",
                "NVS_ENZ_rAChE",
                "NVS_ENZ_rCNOS",
                "NVS_ENZ_rMAOAC",
                "NVS_ENZ_rMAOAP",
                "NVS_ENZ_rMAOBC",
                "NVS_ENZ_rMAOBP",
                "NVS_ENZ_rabI2C",
                "NVS_GPCR_bAdoR_NonSelective",
                "NVS_GPCR_bDR_NonSelective",
                "NVS_GPCR_g5HT4",
                "NVS_GPCR_gH2",
                "NVS_GPCR_gLTB4",
                "NVS_GPCR_gLTD4",
                "NVS_GPCR_gMPeripheral_NonSelective",
                "NVS_GPCR_gOpiateK",
                "NVS_GPCR_h5HT2A",
                "NVS_GPCR_h5HT5A",
                "NVS_GPCR_h5HT6",
                "NVS_GPCR_h5HT7",
                "NVS_GPCR_hAT1",
                "NVS_GPCR_hAdoRA1",
                "NVS_GPCR_hAdoRA2a",
                "NVS_GPCR_hAdra2A",
                "NVS_GPCR_hAdra2C",
                "NVS_GPCR_hAdrb1",
                "NVS_GPCR_hAdrb2",
                "NVS_GPCR_hAdrb3",
                "NVS_GPCR_hDRD1",
                "NVS_GPCR_hDRD2s",
                "NVS_GPCR_hDRD4.4",
                "NVS_GPCR_hH1",
                "NVS_GPCR_hLTB4_BLT1",
                "NVS_GPCR_hM1",
                "NVS_GPCR_hM2",
                "NVS_GPCR_hM3",
                "NVS_GPCR_hM4",
                "NVS_GPCR_hNK2",
                "NVS_GPCR_hOpiate_D1",
                "NVS_GPCR_hOpiate_mu",
                "NVS_GPCR_hTXA2",
                "NVS_GPCR_p5HT2C",
                "NVS_GPCR_r5HT1_NonSelective",
                "NVS_GPCR_r5HT_NonSelective",
                "NVS_GPCR_rAdra1B",
                "NVS_GPCR_rAdra1_NonSelective",
                "NVS_GPCR_rAdra2_NonSelective",
                "NVS_GPCR_rAdrb_NonSelective",
                "NVS_GPCR_rNK1",
                "NVS_GPCR_rNK3",
                "NVS_GPCR_rOpiate_NonSelective",
                "NVS_GPCR_rOpiate_NonSelectiveNa",
                "NVS_GPCR_rSST",
                "NVS_GPCR_rTRH",
                "NVS_GPCR_rV1",
                "NVS_GPCR_rabPAF",
                "NVS_GPCR_rmAdra2B",
                "NVS_IC_hKhERGCh",
                "NVS_IC_rCaBTZCHL",
                "NVS_IC_rCaDHPRCh_L",
                "NVS_IC_rNaCh_site2",
                "NVS_LGIC_bGABARa1",
                "NVS_LGIC_h5HT3",
                "NVS_LGIC_hNNR_NBungSens",
                "NVS_LGIC_rGABAR_NonSelective",
                "NVS_LGIC_rNNR_BungSens",
                "NVS_MP_hPBR",
                "NVS_MP_rPBR",
                "NVS_NR_bER",
                "NVS_NR_bPR",
                "NVS_NR_cAR",
                "NVS_NR_hAR",
                "NVS_NR_hCAR_Antagonist",
                "NVS_NR_hER",
                "NVS_NR_hFXR_Agonist",
                "NVS_NR_hFXR_Antagonist",
                "NVS_NR_hGR",
                "NVS_NR_hPPARa",
                "NVS_NR_hPPARg",
                "NVS_NR_hPR",
                "NVS_NR_hPXR",
                "NVS_NR_hRAR_Antagonist",
                "NVS_NR_hRARa_Agonist",
                "NVS_NR_hTRa_Antagonist",
                "NVS_NR_mERa",
                "NVS_NR_rAR",
                "NVS_NR_rMR",
                "NVS_OR_gSIGMA_NonSelective",
                "NVS_TR_gDAT",
                "NVS_TR_hAdoT",
                "NVS_TR_hDAT",
                "NVS_TR_hNET",
                "NVS_TR_hSERT",
                "NVS_TR_rNET",
                "NVS_TR_rSERT",
                "NVS_TR_rVMAT2",
                "OT_AR_ARELUC_AG_1440",
                "OT_AR_ARSRC1_0480",
                "OT_AR_ARSRC1_0960",
                "OT_ER_ERaERa_0480",
                "OT_ER_ERaERa_1440",
                "OT_ER_ERaERb_0480",
                "OT_ER_ERaERb_1440",
                "OT_ER_ERbERb_0480",
                "OT_ER_ERbERb_1440",
                "OT_ERa_EREGFP_0120",
                "OT_ERa_EREGFP_0480",
                "OT_FXR_FXRSRC1_0480",
                "OT_FXR_FXRSRC1_1440",
                "OT_NURR1_NURR1RXRa_0480",
                "OT_NURR1_NURR1RXRa_1440",
                "TOX21_ARE_BLA_Agonist_ch1",
                "TOX21_ARE_BLA_Agonist_ch2",
                "TOX21_ARE_BLA_agonist_ratio",
                "TOX21_ARE_BLA_agonist_viability",
                "TOX21_AR_BLA_Agonist_ch1",
                "TOX21_AR_BLA_Agonist_ch2",
                "TOX21_AR_BLA_Agonist_ratio",
                "TOX21_AR_BLA_Antagonist_ch1",
                "TOX21_AR_BLA_Antagonist_ch2",
                "TOX21_AR_BLA_Antagonist_ratio",
                "TOX21_AR_BLA_Antagonist_viability",
                "TOX21_AR_LUC_MDAKB2_Agonist",
                "TOX21_AR_LUC_MDAKB2_Antagonist",
                "TOX21_AR_LUC_MDAKB2_Antagonist2",
                "TOX21_AhR_LUC_Agonist",
                "TOX21_Aromatase_Inhibition",
                "TOX21_AutoFluor_HEK293_Cell_blue",
                "TOX21_AutoFluor_HEK293_Media_blue",
                "TOX21_AutoFluor_HEPG2_Cell_blue",
                "TOX21_AutoFluor_HEPG2_Cell_green",
                "TOX21_AutoFluor_HEPG2_Media_blue",
                "TOX21_AutoFluor_HEPG2_Media_green",
                "TOX21_ELG1_LUC_Agonist",
                "TOX21_ERa_BLA_Agonist_ch1",
                "TOX21_ERa_BLA_Agonist_ch2",
                "TOX21_ERa_BLA_Agonist_ratio",
                "TOX21_ERa_BLA_Antagonist_ch1",
                "TOX21_ERa_BLA_Antagonist_ch2",
                "TOX21_ERa_BLA_Antagonist_ratio",
                "TOX21_ERa_BLA_Antagonist_viability",
                "TOX21_ERa_LUC_BG1_Agonist",
                "TOX21_ERa_LUC_BG1_Antagonist",
                "TOX21_ESRE_BLA_ch1",
                "TOX21_ESRE_BLA_ch2",
                "TOX21_ESRE_BLA_ratio",
                "TOX21_ESRE_BLA_viability",
                "TOX21_FXR_BLA_Antagonist_ch1",
                "TOX21_FXR_BLA_Antagonist_ch2",
                "TOX21_FXR_BLA_agonist_ch2",
                "TOX21_FXR_BLA_agonist_ratio",
                "TOX21_FXR_BLA_antagonist_ratio",
                "TOX21_FXR_BLA_antagonist_viability",
                "TOX21_GR_BLA_Agonist_ch1",
                "TOX21_GR_BLA_Agonist_ch2",
                "TOX21_GR_BLA_Agonist_ratio",
                "TOX21_GR_BLA_Antagonist_ch2",
                "TOX21_GR_BLA_Antagonist_ratio",
                "TOX21_GR_BLA_Antagonist_viability",
                "TOX21_HSE_BLA_agonist_ch1",
                "TOX21_HSE_BLA_agonist_ch2",
                "TOX21_HSE_BLA_agonist_ratio",
                "TOX21_HSE_BLA_agonist_viability",
                "TOX21_MMP_ratio_down",
                "TOX21_MMP_ratio_up",
                "TOX21_MMP_viability",
                "TOX21_NFkB_BLA_agonist_ch1",
                "TOX21_NFkB_BLA_agonist_ch2",
                "TOX21_NFkB_BLA_agonist_ratio",
                "TOX21_NFkB_BLA_agonist_viability",
                "TOX21_PPARd_BLA_Agonist_viability",
                "TOX21_PPARd_BLA_Antagonist_ch1",
                "TOX21_PPARd_BLA_agonist_ch1",
                "TOX21_PPARd_BLA_agonist_ch2",
                "TOX21_PPARd_BLA_agonist_ratio",
                "TOX21_PPARd_BLA_antagonist_ratio",
                "TOX21_PPARd_BLA_antagonist_viability",
                "TOX21_PPARg_BLA_Agonist_ch1",
                "TOX21_PPARg_BLA_Agonist_ch2",
                "TOX21_PPARg_BLA_Agonist_ratio",
                "TOX21_PPARg_BLA_Antagonist_ch1",
                "TOX21_PPARg_BLA_antagonist_ratio",
                "TOX21_PPARg_BLA_antagonist_viability",
                "TOX21_TR_LUC_GH3_Agonist",
                "TOX21_TR_LUC_GH3_Antagonist",
                "TOX21_VDR_BLA_Agonist_viability",
                "TOX21_VDR_BLA_Antagonist_ch1",
                "TOX21_VDR_BLA_agonist_ch2",
                "TOX21_VDR_BLA_agonist_ratio",
                "TOX21_VDR_BLA_antagonist_ratio",
                "TOX21_VDR_BLA_antagonist_viability",
                "TOX21_p53_BLA_p1_ch1",
                "TOX21_p53_BLA_p1_ch2",
                "TOX21_p53_BLA_p1_ratio",
                "TOX21_p53_BLA_p1_viability",
                "TOX21_p53_BLA_p2_ch1",
                "TOX21_p53_BLA_p2_ch2",
                "TOX21_p53_BLA_p2_ratio",
                "TOX21_p53_BLA_p2_viability",
                "TOX21_p53_BLA_p3_ch1",
                "TOX21_p53_BLA_p3_ch2",
                "TOX21_p53_BLA_p3_ratio",
                "TOX21_p53_BLA_p3_viability",
                "TOX21_p53_BLA_p4_ch1",
                "TOX21_p53_BLA_p4_ch2",
                "TOX21_p53_BLA_p4_ratio",
                "TOX21_p53_BLA_p4_viability",
                "TOX21_p53_BLA_p5_ch1",
                "TOX21_p53_BLA_p5_ch2",
                "TOX21_p53_BLA_p5_ratio",
                "TOX21_p53_BLA_p5_viability",
                "Tanguay_ZF_120hpf_AXIS_up",
                "Tanguay_ZF_120hpf_ActivityScore",
                "Tanguay_ZF_120hpf_BRAI_up",
                "Tanguay_ZF_120hpf_CFIN_up",
                "Tanguay_ZF_120hpf_CIRC_up",
                "Tanguay_ZF_120hpf_EYE_up",
                "Tanguay_ZF_120hpf_JAW_up",
                "Tanguay_ZF_120hpf_MORT_up",
                "Tanguay_ZF_120hpf_OTIC_up",
                "Tanguay_ZF_120hpf_PE_up",
                "Tanguay_ZF_120hpf_PFIN_up",
                "Tanguay_ZF_120hpf_PIG_up",
                "Tanguay_ZF_120hpf_SNOU_up",
                "Tanguay_ZF_120hpf_SOMI_up",
                "Tanguay_ZF_120hpf_SWIM_up",
                "Tanguay_ZF_120hpf_TRUN_up",
                "Tanguay_ZF_120hpf_TR_up",
                "Tanguay_ZF_120hpf_YSE_up",
            ],
            "delaney_esol": "measured log solubility in mols per litre",
            "freesolv": "y",
            "lipo": "exp",
            "bace_regression": "pIC50",
        }

        if datasets == ["all"]:
            datasets = list(self.datasets_loaded.keys())

        dfs = {}

        from os import path

        if "geognn" in mode:
            base_path = path.join(path.expanduser("~"), f".oce/moleculenet2/")
        else:
            base_path = path.join(path.expanduser("~"), f".oce/moleculenet/")
        for name in datasets:
            dataset_path = path.join(base_path, f"{name}.csv")
            if path.exists(dataset_path):
                print("Loading dataset from local cache... ")
                df = pd.read_csv(dataset_path)
            else:
                print("Downloading dataset... ")
                if not path.exists(base_path):
                    os.mkdir(base_path)
                df = pd.read_csv(self.datasets_loaded[name])
                df.to_csv(dataset_path, index=False)
            if "split" in df.columns:
                df["split"] = df["split"].str.lower()
            dfs[name] = df

        self.dataset_names = datasets
        self.dfs = dfs
        self.metrics = {
            name: "ROC-AUC"
            if self.datasets_types[name] == "classification"
            else "Root Mean Squared Error"
            for name in self.dataset_names
        }
        self.model_names = []
        self.model_params = []
        self.model_database = pd.DataFrame()

        self.file_path = file_path
        if not self.file_path is None and os.path.exists(self.file_path):
            with open(self.file_path, "rb") as f:
                d = pickle.load(f)
                self._load(d["instance_save"])

    """Runs the benchmark dataset on each of the desired models, including datasets with multi-task requirements."""

    def run(self, models: Union[BaseModel, List[BaseModel]]):
        if not isinstance(models, list):
            models = [models]

        for model_ in tqdm(models):
            model = model_.copy()
            print("Running benchmark on model: {}".format(model.name))
            for name, df in tqdm(self.dfs.items()):
                l = []
                print(f"Running dataset {name}")
                property_cols = self.dataset_property_cols[name]
                if isinstance(property_cols, str):
                    property_cols = [property_cols]
                if "small" in self.mode:
                    if len(property_cols) > 5:
                        import random
                        from random import sample

                        random.seed(42)
                        property_cols = sample(property_cols, 5)
                elif "single" in self.mode:
                    if len(property_cols) > 1:
                        import random
                        from random import sample

                        random.seed(42)
                        property_cols = sample(property_cols, 1)

                if "smiles" in df.columns:
                    structure_col = "smiles"
                elif "mol" in df.columns:
                    structure_col = "mol"
                dataset = (
                    oce.BaseDataset(data=df.to_csv(), structure_col=structure_col)
                    + oce.CleanStructures()
                )

                if "geognn" in self.mode:
                    dataset = dataset + oce.gg_ScaffoldSplit(
                        split_proportions=[0.8, 0.1, 0.1]
                    )
                else:
                    dataset = dataset + oce.ScaffoldSplit(
                        split_proportions=[0.8, 0.1, 0.1]
                    )

                for property_col in tqdm(property_cols):
                    print(f"Running property {property_col}")

                    dataset2 = dataset.copy()
                    dataset2.property_col = property_col
                    dataset2 = dataset2 + oce.CleanStructures()

                    model.fit(*dataset2.train_dataset)
                    print(f"Running {model.setting} metrics")
                    if self.eval == "test":
                        results = model.test(*dataset2.test_dataset)
                    else:
                        results = model.test(*dataset2.valid_dataset)

                    l.append(results[self.metrics[name]])

                self.model_database = self.model_database.append(
                    {
                        "Model Name": model.name,
                        "Model Parameters": oce.pretty_params_str(model),
                        "Evaluation": self.eval,
                        "Dataset": name,
                        "Metric": np.mean(l),
                        "Full Metrics List": np.array(l),
                    },
                    ignore_index=True,
                )
                if not self.file_path is None:
                    oce.save(self, self.file_path)

    def _save(self):
        d = {"model_database": self.model_database.to_csv(index=False)}
        return d

    def _load(self, d):
        import io

        self.model_database = pd.read_csv(io.StringIO(d["model_database"]))



from copy import copy

import torch
from torch import nn as nn
from torch.nn import functional as F
from torch_geometric import nn as nng
from torch_geometric.nn import MessagePassing
from torch_scatter import scatter_add
from torch_sparse import SparseTensor, coalesce

from olorenchemengine.internal import mock_imports


class MLP(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.main = [
            nn.Linear(in_dim, 2 * in_dim),
            nn.BatchNorm1d(2 * in_dim),
            nn.ReLU(),
        ]
        self.main.append(nn.Linear(2 * in_dim, out_dim))
        self.main = nn.Sequential(*self.main)

    def forward(self, x):
        return self.main(x)


#################
# Embeddings
#################
class OGBMolEmbedding(nn.Module):
    def __init__(self, dim, embed_edge=True, x_as_list=False):
        super().__init__()
        from ogb.graphproppred.mol_encoder import AtomEncoder, BondEncoder

        self.atom_embedding = AtomEncoder(emb_dim=dim)
        if embed_edge:
            self.edge_embedding = BondEncoder(emb_dim=dim)
        self.x_as_list = x_as_list

    def forward(self, data):
        data = new(data)
        # try:
        data.x = self.atom_embedding(data.x)
        if self.x_as_list:
            data.x = [data.x]
        if hasattr(self, "edge_embedding"):
            data.edge_attr = self.edge_embedding(data.edge_attr)
        return data
        # except Exception as e:
        #    print(e)
        #    print(data.smiles)


class NodeEmbedding(nn.Module):
    def __init__(self, in_dim, out_dim, x_as_list=False):
        super().__init__()
        self.embedding = nn.Embedding(num_embeddings=in_dim, embedding_dim=out_dim)
        self.x_as_list = x_as_list

    def forward(self, data):
        data = new(data)
        data.x = self.embedding(data.x)
        if self.x_as_list:
            data.x = [data.x]
        return data


class VNAgg(nn.Module):
    def __init__(self, dim, conv_type="gin"):
        super().__init__()
        self.conv_type = conv_type
        if "gin" in conv_type:
            self.mlp = nn.Sequential(MLP(dim, dim), nn.BatchNorm1d(dim), nn.ReLU())
        elif "gcn" in conv_type:
            self.W0 = nn.Linear(dim, dim)
            self.W1 = nn.Linear(dim, dim)
            self.nl_bn = nn.Sequential(nn.BatchNorm1d(dim), nn.ReLU())
        else:
            raise NotImplementedError("Unrecognised model conv : {}".format(conv_type))

    def forward(self, virtual_node, embeddings, batch_vector):
        if batch_vector.size(0) > 0:  # ...or the operation will crash for empty graphs
            from torch_geometric.nn import global_add_pool

            G = global_add_pool(embeddings, batch_vector)
        else:
            G = torch.zeros_like(virtual_node)
        if "gin" in self.conv_type:
            virtual_node = virtual_node + G
            virtual_node = self.mlp(virtual_node)
        elif "gcn" in self.conv_type:
            virtual_node = self.W0(virtual_node) + self.W1(G)
            virtual_node = self.nl_bn(virtual_node)
        else:
            raise NotImplementedError(
                "Unrecognised model conv : {}".format(self.conv_type)
            )
        return virtual_node


class ConvBlock(nn.Module):
    def __init__(
        self,
        dim,
        dropout=0.5,
        activation=F.relu,
        virtual_node=False,
        virtual_node_agg=True,
        k=4,
        last_layer=False,
        conv_type="gin",
        edge_embedding=None,
    ):
        super().__init__()
        self.edge_embed = edge_embedding
        self.conv_type = conv_type
        if conv_type == "gin+":
            self.conv = GINEPLUS(MLP(dim, dim), dim, k=k)
        elif conv_type == "naivegin+":
            self.conv = NAIVEGINEPLUS(MLP(dim, dim), dim, k=k)
        elif conv_type == "gin":
            from torch_geometric.nn import GINEConv

            self.conv = GINEConv(MLP(dim, dim), train_eps=True)
        elif conv_type == "gcn":
            from torch_geometric.nn import GCNConv

            self.conv = GCNConv(dim, dim)
        self.norm = nn.BatchNorm1d(dim)
        self.act = activation or nn.Identity()
        self.last_layer = last_layer

        self.dropout_ratio = dropout

        self.virtual_node = virtual_node
        self.virtual_node_agg = virtual_node_agg
        if self.virtual_node and self.virtual_node_agg:
            self.vn_aggregator = VNAgg(dim, conv_type=conv_type)

    def forward(self, data):
        data = new(data)
        x, ei, ea, b = data.x, data.edge_index, data.edge_attr, data.batch
        mhei, d = data.multihop_edge_index, data.distance
        h = x
        if self.virtual_node:
            if self.conv_type == "gin+":
                h[0] = h[0] + data.virtual_node[b]
            else:
                h = h + data.virtual_node[b]
        if self.conv_type == "gin":
            H = self.conv(h, ei, edge_attr=self.edge_embed(ea))
        elif self.conv_type == "gcn":
            H = self.conv(h, ei)
        else:
            H = self.conv(h, mhei, d, self.edge_embed(ea))
        if self.conv_type == "gin+":
            h = H[0]
        else:
            h = H
        h = self.norm(h)
        if not self.last_layer:
            h = self.act(h)
        h = F.dropout(h, self.dropout_ratio, training=self.training)

        if self.virtual_node and self.virtual_node_agg:
            v = self.vn_aggregator(data.virtual_node, h, b)
            v = F.dropout(v, self.dropout_ratio, training=self.training)
            data.virtual_node = v
        if self.conv_type == "gin+":
            H[0] = h
            h = H
        data.x = h
        return data


class GlobalPool(nn.Module):
    def __init__(self, fun, cat_size=False, cat_candidates=False):
        super().__init__()
        self.cat_size = cat_size
        from torch_geometric.nn import (
            global_add_pool,
            global_max_pool,
            global_mean_pool,
        )

        if fun.lower() == "add":
            self.fun = global_add_pool
        elif fun.lower() == "mean":
            self.fun = global_mean_pool
        elif fun.lower() == "max":
            self.fun = global_max_pool
        self.cat_candidates = cat_candidates

    def forward(self, batch):
        x, b = batch.x, batch.batch
        pooled = self.fun(x, b, size=batch.num_graphs)
        from torch_geometric.nn import (
            global_add_pool,
            global_max_pool,
            global_mean_pool,
        )

        if self.cat_size:
            sizes = global_add_pool(
                torch.ones(x.size(0), 1).type_as(x), b, size=batch.num_graphs
            )
            pooled = torch.cat([pooled, sizes], dim=1)
        if self.cat_candidates:
            ei = batch.edge_index
            mask = batch.edge_attr == 3
            candidates = scatter_add(
                x[ei[0, mask]], b[ei[0, mask]], dim=0, dim_size=batch.num_graphs
            )
            pooled = torch.cat([pooled, candidates], dim=1)
        return pooled

class GINNetwork(nn.Module):
    def __init__(
        self,
        hidden=100,
        out_dim=128,
        layers=3,
        dropout=0.5,
        virtual_node=False,
        k=4,
        conv_type="gin+",
    ):
        super().__init__()
        self.k = k
        self.conv_type = conv_type
        from ogb.graphproppred.mol_encoder import AtomEncoder, BondEncoder

        convs = [
            ConvBlock(
                hidden,
                dropout=dropout,
                virtual_node=virtual_node,
                k=min(i + 1, k),
                conv_type=conv_type,
                edge_embedding=BondEncoder(emb_dim=hidden),
            )
            for i in range(layers - 1)
        ]
        convs.append(
            ConvBlock(
                hidden,
                dropout=dropout,
                virtual_node=virtual_node,
                virtual_node_agg=False,  # on last layer, use but do not update virtual node
                last_layer=True,
                k=min(layers, k),
                conv_type=conv_type,
                edge_embedding=BondEncoder(emb_dim=hidden),
            )
        )
        self.main = nn.Sequential(
            OGBMolEmbedding(hidden, embed_edge=False, x_as_list=(conv_type == "gin+")),
            *convs
        )
        self.aggregate = nn.Sequential(GlobalPool("mean"), nn.Linear(hidden, out_dim))
        self.virtual_node = virtual_node
        if self.virtual_node:
            self.v0 = nn.Parameter(torch.zeros(1, hidden), requires_grad=True)

    def forward(self, data):
        data = make_multihop_edges(data, self.k)
        if self.virtual_node:
            data.virtual_node = self.v0.expand(data.num_graphs, self.v0.shape[-1])
        g = self.main(data)
        if self.conv_type == "gin+":
            g.x = g.x[0]
        return self.aggregate(g)


def make_multihop_edges(data, k):
    """
    Adds edges corresponding to distances up to k to a data object.
    :param data: torch_geometric.data object, in coo format
    (ie an edge (i, j) with label v is stored with an arbitrary index u as:
     edge_index[0, u] = i, edge_index[1, u]=j, edge_attr[u]=v)
    :return: a new data object with new fields, multihop_edge_index and distance.
    distance[u] contains values from 1 to k corresponding to the distance between
    multihop_edge_index[0, u] and multihop_edge_index[1, u]
    """
    data = new(data)

    N = data.num_nodes
    E = data.num_edges
    if E == 0:
        data.multihop_edge_index = torch.empty_like(data.edge_index)
        data.distance = torch.empty_like(data.multihop_edge_index[0])
        return data

    # Get the distance 0
    multihop_edge_index = torch.arange(
        0, N, dtype=data.edge_index[0].dtype, device=data.x.device
    )
    distance = torch.zeros_like(multihop_edge_index)
    multihop_edge_index = multihop_edge_index.unsqueeze(0).repeat(2, 1)

    # Get the distance 1
    multihop_edge_index = torch.cat((multihop_edge_index, data.edge_index), dim=1)
    distance = torch.cat((distance, torch.ones_like(data.edge_index[0])), dim=0)

    A = SparseTensor(
        row=data.edge_index[0],
        col=data.edge_index[1],
        value=torch.ones_like(data.edge_index[0]).float(),
        sparse_sizes=(N, N),
        is_sorted=False,
    )
    Ad = A  # A to the power d

    # Get larger distances
    from torch_sparse import matmul

    for d in range(2, k + 1):
        Ad = matmul(Ad, A)
        row, col, v = Ad.coo()
        d_edge_index = torch.stack((row, col))
        d_edge_attr = torch.empty_like(row).fill_(d)
        multihop_edge_index = torch.cat((multihop_edge_index, d_edge_index), dim=1)
        distance = torch.cat((distance, d_edge_attr), dim=0)

    # remove dupicate, keep only shortest distance
    multihop_edge_index, distance = coalesce(
        multihop_edge_index, distance, N, N, op="min"
    )

    data.multihop_edge_index = multihop_edge_index
    data.distance = distance

    return data


class NAIVEGINEPLUS(MessagePassing):
    def __init__(self, fun, dim, k=4, **kwargs):
        super().__init__(aggr="add")
        self.k = k
        self.nn = fun
        self.eps = nn.Parameter(torch.zeros(k + 1, dim), requires_grad=True)

    def forward(self, x, multihop_edge_index, distance, edge_attr):
        assert x.size(-1) == edge_attr.size(-1)
        result = (1 + self.eps[0]) * x
        for i in range(self.k):
            if i == 0:
                out = self.propagate(
                    multihop_edge_index[:, distance == i + 1], edge_attr=edge_attr, x=x
                )
            else:
                out = self.propagate(
                    multihop_edge_index[:, distance == i + 1], edge_attr=None, x=x
                )
            result += (1 + self.eps[i + 1]) * out
        result = self.nn(result)
        return result

    def message(self, x_j, edge_attr):
        if edge_attr is not None:
            return F.relu(x_j + edge_attr)
        else:
            return F.relu(x_j)

    def __repr__(self):
        return "{}(nn={}, k={})".format(
            self.__class__.__name__, self.nn, self.eps.size(0)
        )


class GINEPLUS(MessagePassing):
    def __init__(self, fun, dim, k=4, **kwargs):
        super().__init__(aggr="add")
        self.k = k
        self.nn = fun
        self.eps = nn.Parameter(torch.zeros(k + 1, dim), requires_grad=True)

    def forward(self, XX, multihop_edge_index, distance, edge_attr):
        """Warning, XX is now a list of previous xs, with x[0] being the last layer"""
        assert len(XX) >= self.k
        assert XX[-1].size(-1) == edge_attr.size(-1)
        result = (1 + self.eps[0]) * XX[0]
        for i, x in enumerate(XX):
            if i >= self.k:
                break
            if i == 0:
                out = self.propagate(
                    multihop_edge_index[:, distance == i + 1], edge_attr=edge_attr, x=x
                )
            else:
                out = self.propagate(
                    multihop_edge_index[:, distance == i + 1], edge_attr=None, x=x
                )
            result += (1 + self.eps[i + 1]) * out
        result = self.nn(result)
        return [result] + XX

    def message(self, x_j, edge_attr):
        if edge_attr is not None:
            return F.relu(x_j + edge_attr)
        else:
            return F.relu(x_j)


def new(data):
    return copy(data)



"""
BSD 3-Clause License

Copyright (c) 2017, Mol2vec developers
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
Features - Main Mol2vec Module
==============================
w. PR for gensim 4.0 from tonyreina and samoturk
"""
import timeit

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from rdkit import Chem
from rdkit.Chem import AllChem, PandasTools
from tqdm import tqdm


class DfVec(object):
    """
    Helper class to store vectors in a pandas DataFrame

    Parameters
    ----------
    vec: np.array
    """

    def __init__(self, vec):
        self.vec = vec
        if type(self.vec) != np.ndarray:
            raise TypeError("numpy.ndarray expected, got %s" % type(self.vec))

    def __str__(self):
        return "%s dimensional vector" % str(self.vec.shape)

    __repr__ = __str__

    def __len__(self):
        return len(self.vec)

    _repr_html_ = __str__


class MolSentence:
    """Class for storing mol sentences in pandas DataFrame"""

    def __init__(self, sentence):
        self.sentence = sentence
        if type(self.sentence[0]) != str:
            raise TypeError("List with strings expected")

    def __len__(self):
        return len(self.sentence)

    def __str__(self):  # String representation
        return "MolSentence with %i words" % len(self.sentence)

    __repr__ = __str__  # Default representation

    def contains(self, word):
        """Contains (and __contains__) method enables usage of "'Word' in MolSentence"""
        if word in self.sentence:
            return True
        else:
            return False

    __contains__ = contains  # MolSentence.contains('word')

    def __iter__(self):  # Iterate over words (for word in MolSentence:...)
        for x in self.sentence:
            yield x

    _repr_html_ = __str__


def mol2sentence(mol, radius):

    """Calculates ECFP (Morgan fingerprint) and returns identifiers of substructures as 'sentence' (string).
    Returns a tuple with 1) a list with sentence for each radius and 2) a sentence with identifiers from all radii
    combined.
    NOTE: Words are ALWAYS reordered according to atom order in the input mol object.
    NOTE: Due to the way how Morgan FPs are generated, number of identifiers at each radius is smaller

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
    radius : float
        Fingerprint radius
    Returns
    -------
    identifier sentence
        List with sentences for each radius
    alternating sentence
        Sentence (list) with identifiers from all radii combined
    """
    radii = list(range(int(radius) + 1))
    info = {}
    _ = AllChem.GetMorganFingerprint(
        mol, radius, bitInfo=info
    )  # info: dictionary identifier, atom_idx, radius

    mol_atoms = [a.GetIdx() for a in mol.GetAtoms()]
    dict_atoms = {x: {r: None for r in radii} for x in mol_atoms}

    for element in info:
        for atom_idx, radius_at in info[element]:
            dict_atoms[atom_idx][
                radius_at
            ] = element  # {atom number: {fp radius: identifier}}

    # iterate over all atoms and radii
    identifier_sentences = []

    for r in radii:  # iterate over radii to get one sentence per radius
        identifiers = []
        for atom in dict_atoms:  # iterate over atoms
            # get one sentence per radius
            identifiers.append(dict_atoms[atom][r])
        identifier_sentences.append(list(map(str, [x for x in identifiers if x])))

    # merge identifiers alternating radius to sentence: atom 0 radius0, atom 0 radius 1, etc.
    identifiers_alt = []
    for atom in dict_atoms:  # iterate over atoms
        for r in radii:  # iterate over radii
            identifiers_alt.append(dict_atoms[atom][r])

    alternating_sentence = map(str, [x for x in identifiers_alt if x])

    return list(identifier_sentences), list(alternating_sentence)


def mol2alt_sentence(mol, radius):
    """Same as mol2sentence() expect it only returns the alternating sentence
    Calculates ECFP (Morgan fingerprint) and returns identifiers of substructures as 'sentence' (string).
    Returns a tuple with 1) a list with sentence for each radius and 2) a sentence with identifiers from all radii
    combined.
    NOTE: Words are ALWAYS reordered according to atom order in the input mol object.
    NOTE: Due to the way how Morgan FPs are generated, number of identifiers at each radius is smaller

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
    radius : float
        Fingerprint radius

    Returns
    -------
    list
        alternating sentence
    combined
    """
    radii = list(range(int(radius) + 1))
    info = {}
    _ = AllChem.GetMorganFingerprint(
        mol, radius, bitInfo=info
    )  # info: dictionary identifier, atom_idx, radius

    mol_atoms = [a.GetIdx() for a in mol.GetAtoms()]
    dict_atoms = {x: {r: None for r in radii} for x in mol_atoms}

    for element in info:
        for atom_idx, radius_at in info[element]:
            dict_atoms[atom_idx][
                radius_at
            ] = element  # {atom number: {fp radius: identifier}}

    # merge identifiers alternating radius to sentence: atom 0 radius0, atom 0 radius 1, etc.
    identifiers_alt = []
    for atom in dict_atoms:  # iterate over atoms
        for r in radii:  # iterate over radii
            identifiers_alt.append(dict_atoms[atom][r])

    alternating_sentence = map(str, [x for x in identifiers_alt if x])

    return list(alternating_sentence)


def _parallel_job(mol, r):
    """Helper function for joblib jobs"""
    if mol is not None:
        smiles = Chem.MolToSmiles(mol)
        mol = Chem.MolFromSmiles(smiles)
        sentence = mol2alt_sentence(mol, r)
        return " ".join(sentence)


def _read_smi(file_name):
    while True:
        line = file_name.readline()
        if not line:
            break
        yield Chem.MolFromSmiles(line.split("\t")[0])


def generate_corpus(in_file, out_file, r, sentence_type="alt", n_jobs=1):

    """Generates corpus file from sdf

    Parameters
    ----------
    in_file : str
        Input sdf
    out_file : str
        Outfile name prefix, suffix is either _r0, _r1, etc. or _alt_r1 (max radius in alt sentence)
    r : int
        Radius of morgan fingerprint
    sentence_type : str
        Options:    'all' - generates all corpus files for all types of sentences,
                    'alt' - generates a corpus file with only combined alternating sentence,
                    'individual' - generates corpus files for each radius
    n_jobs : int
        Number of cores to use (only 'alt' sentence type is parallelized)
    Returns
    -------
    """

    # File type detection
    in_split = in_file.split(".")
    if in_split[-1].lower() not in ["sdf", "smi", "ism", "gz"]:
        raise ValueError("File extension not supported (sdf, smi, ism, sdf.gz, smi.gz)")
    gzipped = False
    if in_split[-1].lower() == "gz":
        gzipped = True
        if in_split[-2].lower() not in ["sdf", "smi", "ism"]:
            raise ValueError(
                "File extension not supported (sdf, smi, ism, sdf.gz, smi.gz)"
            )

    file_handles = []

    # write only files which contain corpus
    if (sentence_type == "individual") or (sentence_type == "all"):

        f1 = open(out_file + "_r0.corpus", "w")
        f2 = open(out_file + "_r1.corpus", "w")
        file_handles.append(f1)
        file_handles.append(f2)

    if (sentence_type == "alt") or (sentence_type == "all"):
        f3 = open(out_file, "w")
        file_handles.append(f3)

    if gzipped:
        import gzip

        if in_split[-2].lower() == "sdf":
            mols_file = gzip.open(in_file, mode="r")
            suppl = Chem.ForwardSDMolSupplier(mols_file)
        else:
            mols_file = gzip.open(in_file, mode="rt")
            suppl = _read_smi(mols_file)
    else:
        if in_split[-1].lower() == "sdf":
            suppl = Chem.ForwardSDMolSupplier(in_file)
        else:
            mols_file = open(in_file, mode="rt")
            suppl = _read_smi(mols_file)

    if sentence_type == "alt":  # This can run parallelized
        result = Parallel(n_jobs=n_jobs, verbose=1)(
            delayed(_parallel_job)(mol, r) for mol in suppl
        )
        for i, line in enumerate(result):
            f3.write(str(line) + "\n")
        print("% molecules successfully processed.")

    else:
        for mol in suppl:
            if mol is not None:
                smiles = Chem.MolToSmiles(mol)
                mol = Chem.MolFromSmiles(smiles)
                identifier_sentences, alternating_sentence = mol2sentence(mol, r)

                identifier_sentence_r0 = " ".join(identifier_sentences[0])
                identifier_sentence_r1 = " ".join(identifier_sentences[1])
                alternating_sentence_r0r1 = " ".join(alternating_sentence)

                if len(smiles) != 0:
                    if (sentence_type == "individual") or (sentence_type == "all"):
                        f1.write(str(identifier_sentence_r0) + "\n")
                        f2.write(str(identifier_sentence_r1) + "\n")

                    if (sentence_type == "alt") or (sentence_type == "all"):
                        f3.write(str(alternating_sentence_r0r1) + "\n")

    for fh in file_handles:
        fh.close()


def _read_corpus(file_name):
    while True:
        line = file_name.readline()
        if not line:
            break
        yield line.split()


def insert_unk(corpus, out_corpus, threshold=3, uncommon="UNK"):
    """Handling of uncommon "words" (i.e. identifiers). It finds all least common identifiers (defined by threshold) and
    replaces them by 'uncommon' string.
    Parameters
    ----------
    corpus : str
        Input corpus file
    out_corpus : str
        Outfile corpus file
    threshold : int
        Number of identifier occurrences to consider it uncommon
    uncommon : str
        String to use to replace uncommon words/identifiers
    Returns
    -------
    """
    # Find least common identifiers in corpus
    f = open(corpus)
    unique = {}
    for i, x in tqdm(enumerate(_read_corpus(f)), desc="Counting identifiers in corpus"):
        for identifier in x:
            if identifier not in unique:
                unique[identifier] = 1
            else:
                unique[identifier] += 1
    n_lines = i + 1
    least_common = set([x for x in unique if unique[x] <= threshold])
    f.close()

    f = open(corpus)
    fw = open(out_corpus, mode="w")
    for line in tqdm(_read_corpus(f), total=n_lines, desc="Inserting %s" % uncommon):
        intersection = set(line) & least_common
        if len(intersection) > 0:
            new_line = []
            for item in line:
                if item in least_common:
                    new_line.append(uncommon)
                else:
                    new_line.append(item)
            fw.write(" ".join(new_line) + "\n")
        else:
            fw.write(" ".join(line) + "\n")
    f.close()
    fw.close()


def train_word2vec_model(
    infile_name,
    outfile_name=None,
    vector_size=100,
    window=10,
    min_count=3,
    n_jobs=1,
    method="skip-gram",
    **kwargs
):
    """Trains word2vec (Mol2vec, ProtVec) model on corpus file extracted from molecule/protein sequences.
    The corpus file is treated as LineSentence corpus (one sentence = one line, words separated by whitespaces)

    Parameters
    ----------
    infile_name : str
        Corpus file, e.g. proteins split in n-grams or compound identifier
    outfile_name : str
        Name of output file where word2vec model should be saved
    vector_size : int
        Number of dimensions of vector
    window : int
        Number of words considered as context
    min_count : int
        Number of occurrences a word should have to be considered in training
    n_jobs : int
        Number of cpu cores used for calculation
    method : str
        Method to use in model training. Options cbow and skip-gram, default: skip-gram)

    Returns
    -------
    word2vec.Word2Vec
    """
    if method.lower() == "skip-gram":
        sg = 1
    elif method.lower() == "cbow":
        sg = 0
    else:
        raise ValueError("skip-gram or cbow are only valid options")

    start = timeit.default_timer()

    from gensim.models import word2vec

    corpus = word2vec.LineSentence(infile_name)
    model = word2vec.Word2Vec(
        corpus,
        size=vector_size,
        window=window,
        min_count=min_count,
        workers=n_jobs,
        sg=sg,
        **kwargs
    )
    if outfile_name:
        model.save(outfile_name)

    stop = timeit.default_timer()
    print("Runtime: ", round((stop - start) / 60, 2), " minutes")
    return model


def remove_salts_solvents(smiles, hac=3):
    """Remove solvents and ions have max 'hac' heavy atoms. This function removes any fragment in molecule that has
    number of heavy atoms <= "hac" and it might not be an actual solvent or salt

    Parameters
    ----------
    smiles : str
        SMILES
    hac : int
        Max number of heavy atoms
    Returns
    -------
    str
        smiles
    """
    save = []
    for el in smiles.split("."):
        mol = Chem.MolFromSmiles(str(el))
        if mol.GetNumHeavyAtoms() <= hac:
            save.append(mol)

    return ".".join([Chem.MolToSmiles(x) for x in save])


def sentences2vec(sentences, model, unseen=None):
    """Generate vectors for each sentence (list) in a list of sentences. Vector is simply a
    sum of vectors for individual words.

    Parameters
    ----------
    sentences : list, array
        List with sentences
    model : word2vec.Word2Vec
        Gensim word2vec model
    unseen : None, str
        Keyword for unseen words. If None, those words are skipped.
        https://stats.stackexchange.com/questions/163005/how-to-set-the-dictionary-for-text-analysis-using-neural-networks/163032#163032
    Returns
    -------
    np.array
    """
    import gensim

    GENSIM4 = gensim.__version__ > "4"  # GENSIM4 has some API changes from GENSIM 3
    if GENSIM4:
        keys = set(model.wv.key_to_index)
    else:
        keys = set(model.wv.vocab.keys())
    vec = []
    if unseen:
        unseen_vec = model.wv.word_vec(unseen)

    for sentence in sentences:
        if unseen:
            vec.append(
                sum(
                    [
                        model.wv.word_vec(y)
                        if y in set(sentence) & keys
                        else unseen_vec
                        for y in sentence
                    ]
                )
            )
        else:
            vec.append(
                sum(
                    [
                        model.wv.word_vec(y)
                        for y in sentence
                        if y in set(sentence) & keys
                    ]
                )
            )
    return np.array(vec)


def featurize(in_file, out_file, model_path, r, uncommon=None):
    """Featurize mols in SDF, SMI.
    SMILES are regenerated with RDKit to get canonical SMILES without chirality information.
    Parameters
    ----------
    in_file : str
        Input SDF, SMI, ISM (or GZ)
    out_file : str
        Output csv
    model_path : str
        File path to pre-trained Gensim word2vec model
    r : int
        Radius of morgan fingerprint
    uncommon : str
        String to used to replace uncommon words/identifiers while training. Vector obtained for 'uncommon' will be used
        to encode new (unseen) identifiers
    Returns
    -------
    """
    # Load the model
    from gensim.models import word2vec

    word2vec_model = word2vec.Word2Vec.load(model_path)
    if uncommon:
        try:
            word2vec_model[uncommon]
        except KeyError:
            raise KeyError(
                "Selected word for uncommon: %s not in vocabulary" % uncommon
            )

    # File type detection
    in_split = in_file.split(".")
    f_type = in_split[-1].lower()
    if f_type not in ["sdf", "smi", "ism", "gz"]:
        raise ValueError("File extension not supported (sdf, smi, ism, sdf.gz, smi.gz)")
    if f_type == "gz":
        if in_split[-2].lower() not in ["sdf", "smi", "ism"]:
            raise ValueError(
                "File extension not supported (sdf, smi, ism, sdf.gz, smi.gz)"
            )
        else:
            f_type = in_split[-2].lower()

    print("Loading molecules.")
    if f_type == "sdf":
        df = PandasTools.LoadSDF(in_file)
        print("Keeping only molecules that can be processed by RDKit.")
        df = df[df["ROMol"].notnull()]
        df["Smiles"] = df["ROMol"].map(Chem.MolToSmiles)
    else:
        df = pd.read_csv(
            in_file, delimiter="\t", usecols=[0, 1], names=["Smiles", "ID"]
        )  # Assume <tab> separated
        PandasTools.AddMoleculeColumnToFrame(df, smilesCol="Smiles")
        print("Keeping only molecules that can be processed by RDKit.")
        df = df[df["ROMol"].notnull()]
        df["Smiles"] = df["ROMol"].map(Chem.MolToSmiles)  # Recreate SMILES

    print("Featurizing molecules.")
    df["mol-sentence"] = df.apply(
        lambda x: MolSentence(mol2alt_sentence(x["ROMol"], r)), axis=1
    )
    vectors = sentences2vec(df["mol-sentence"], word2vec_model, unseen=uncommon)
    df_vec = pd.DataFrame(
        vectors, columns=["mol2vec-%03i" % x for x in range(vectors.shape[1])]
    )
    df_vec.index = df.index
    df = df.join(df_vec)

    df.drop(["ROMol", "mol-sentence"], axis=1).to_csv(out_file)



"""
Visualize interpret is for visualizations that help explain why a model makes a
certain prediction. 
"""

from rdkit import Chem
from tqdm import tqdm

from olorenchemengine.base_class import *
from olorenchemengine.dataset import *
from olorenchemengine.internal import *
from olorenchemengine.interpret import *
from olorenchemengine.manager import *
from olorenchemengine.representations import *
from olorenchemengine.uncertainty import *
from olorenchemengine.visualizations import *


class VisualizePredictionSensitivity(BaseVisualization):
    """
    Visualize the sensitivity of a model's prediction to atom-level perturbations
    generated by the SwapMutations perturbations engine.

    Parameters:
        model (BaseModel): The model to interpret.
        query_compound (str): The smiles string of the compound to interpret.
        radius (int): The radius of the perturbations to generate with SwapMutations
        n (int): The number of perturbations to generate centerred aroudn each atom.
    """

    @log_arguments
    def __init__(
        self,
        model: BaseModel,
        query_compound: str,
        radius: int = 2,
        n: int = 200,
        colorscale="viridis",
        bottom_quantile=0.75,
        top_quantile=0.95,
        nbins=3,
        log=True,
        **kwargs
    ):

        super().__init__(log=False, **kwargs)
        self.packages = ["olorenrenderer"]

        self.mutator = SwapMutations(radius=radius)
        self.model = model
        self.colorscale = colorscale
        self.bottom_quantile = bottom_quantile
        self.top_quantile = top_quantile
        self.nbins = nbins

        self.smiles = query_compound
        self.mol = Chem.MolFromSmiles(self.smiles)

        vals = []
        for i, a in tqdm(enumerate(self.mol.GetAtoms())):
            smiles_list = []
            for i in range(n):
                smiles = self.mutator.get_compound_at_idx(
                    Chem.Mol(self.mol), a.GetIdx()
                )
                if smiles is None or Chem.MolFromSmiles(smiles) is None:
                    continue
                else:
                    smiles_list.append(smiles)
            if len(smiles_list) > 3:
                preds = self.model.predict(smiles_list)
                stdev = np.std(preds)
                a.SetDoubleProp("stdev", stdev)
                vals.append(stdev)
            else:
                print("Not enough perturbations for atom", a.GetIdx())

        bottom_threshold = np.quantile(vals, self.bottom_quantile)
        top_threshold = np.quantile(vals, self.top_quantile)

        for a in self.mol.GetAtoms():
            if a.HasProp("stdev"):
                val = a.GetDoubleProp("stdev")
                if val <= bottom_threshold:
                    pass
                elif val >= top_threshold:
                    a.SetDoubleProp("bin", 1)
                    a.SetAtomMapNum(self.nbins)
                else:
                    normalized_val = (val - bottom_threshold) / (
                        top_threshold - bottom_threshold
                    )
                    bin_number = np.around(normalized_val * self.nbins)
                    if int(bin_number) > 0:
                        a.SetAtomMapNum(int(bin_number))

    def get_data(self):

        def rgb_to_hex(x, colorscale=self.colorscale):
            if not isinstance(x, list):
                x = [x]
                
            from plotly.colors import sample_colorscale
            
            x = sample_colorscale(colorscale, x, colortype="hex")
            x = np.rint(np.array(x) * 255).astype(int)
            return ["#%02x%02x%02x" % (x_[0], x_[1], x_[2]) for x_ in x]

        return {
            "SMILES": Chem.MolToSmiles(self.mol),
            "highlights": [
                [i + 1, rgb_to_hex((i + 1) / self.nbins)[0]] for i in range(self.nbins)
            ],
        }



"""
Contains the basic framework for hyperparameter optimization.

We use hyperopt as our framework for hyperparameter optimization, and the class Opt
functions as the bridge between olorenchemengine and hyperopt. Hyperparameters are defined
in Opt which is used as an argument in a BaseClass object's instantiation. These
hyperparameters are then collated and used for hyperparameter optimization.

The following is a brief introduction to hyperopt and is a useful starting point for
understanding our hyperparameter optimization engine:
https://github.com/hyperopt/hyperopt/wiki/FMin.
"""

from abc import abstractproperty
from typing import *

from hyperopt import fmin, hp, tpe
from hyperopt.pyll import scope

from .base_class import *
from .benchmarks import *
from .manager import *


class Opt(BaseClass):
    @abstractproperty
    def get_hp(self):
        pass

    @log_arguments
    def __init__(self, label, *args, use_int=False, **kwargs):
        self.label = label
        if use_int:
            self.hp = scope.int(self.get_hp(label, *args, **kwargs))
        else:
            self.hp = self.get_hp(label, *args, **kwargs)

    def _load(self, d):
        pass

    def _save(self):
        return {}


class OptChoice(Opt):
    @property
    def get_hp(self):
        return hp.choice


class OptRandInt(Opt):
    @property
    def get_hp(self):
        return hp.randint


class OptUniform(Opt):
    @property
    def get_hp(self):
        return hp.uniform


class OptQUniform(Opt):
    @property
    def get_hp(self):
        return hp.quniform


class OptLogUniform(Opt):
    @property
    def get_hp(self):
        return hp.loguniform


class OptQLogUniform(Opt):
    @property
    def get_hp(self):
        return hp.qloguniform


class OptQNormal(Opt):
    @property
    def get_hp(self):
        return hp.qnormal


class OptLogNormal(Opt):
    @property
    def get_hp(self):
        return hp.lognormal


class OptQLogNormal(Opt):
    @property
    def get_hp(self):
        return hp.qnormal


def index_hyperparameters(object: BaseClass) -> dict:
    """
    Returns a dictionary of hyperparameters for the model.
    """
    if issubclass(type(object), Opt):
        d = {object.label: object.hp}
        for arg in object.args:
            d.update(index_hyperparameters(arg))
        return d
    if issubclass(type(object), BaseClass):
        d = {}
        for arg in object.args:
            count = len(d)
            d_ = index_hyperparameters(arg)
            d.update(d_)
            if not len(d) == count + len(d_):
                raise Exception(
                    f"Hyperparameter indexing failed, overlapping keys in {d_}"
                )
        for k, v in object.kwargs.items():
            count = len(d)
            d_ = index_hyperparameters(v)
            d.update(d_)
            if not len(d) == count + len(d_):
                raise Exception(
                    f"Hyperparameter indexing failed, overlapping keys in {d_}"
                )
        return d
    elif issubclass(type(object), dict):
        d = {}
        for arg in object["args"]:
            count = len(d)
            d_ = index_hyperparameters(arg)
            d.update(d_)
            if not len(d) == count + len(d_):
                raise Exception(
                    f"Hyperparameter indexing failed, overlapping keys in {d_}"
                )
        for k, v in object["kwargs"].items():
            count = len(d)
            d_ = index_hyperparameters(v)
            d.update(d_)
            if not len(d) == count + len(d_):
                raise Exception(
                    f"Hyperparameter indexing failed, overlapping keys in {d_}"
                )
        return d
    elif issubclass(type(object), list):
        d = {}
        for x in object:
            count = len(d)
            d_ = index_hyperparameters(x)
            d.update(d_)
            if not len(d) == count + len(d_):
                raise Exception(
                    f"Hyperparameter indexing failed, overlapping keys in {d_}"
                )
        return d
    elif (
        object is None
        or issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
    ):
        return {}
    else:
        print(object)
        raise ValueError


def load_hyperparameters_(object: BaseClass, hyperparameter_dictionary: dict) -> dict:
    if issubclass(type(object), Opt):
        return load_hyperparameters_(
            hyperparameter_dictionary[object.label], hyperparameter_dictionary
        )

    if issubclass(type(object), BaseClass):
        return {
            **{"BC_class_name": type(object).__name__},
            **{
                "args": [
                    load_hyperparameters_(arg, hyperparameter_dictionary)
                    for arg in object.args
                ]
            },
            **{
                "kwargs": {
                    k: load_hyperparameters_(v, hyperparameter_dictionary)
                    for k, v in object.kwargs.items()
                }
            },
        }
    elif issubclass(type(object), dict):
        return {
            **{"BC_class_name": object["BC_class_name"]},
            **{
                "args": [
                    load_hyperparameters_(arg, hyperparameter_dictionary)
                    for arg in object["args"]
                ]
            },
            **{
                "kwargs": {
                    k: load_hyperparameters_(v, hyperparameter_dictionary)
                    for k, v in object["kwargs"].items()
                }
            },
        }
    elif (
        object is None
        or issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
    ):
        return object
    elif issubclass(type(object), list):
        return [load_hyperparameters_(x, hyperparameter_dictionary) for x in object]
    else:
        print(object)
        raise ValueError


def load_hyperparameters(object: BaseClass, hyperparameter_dictionary: dict) -> dict:
    mp = load_hyperparameters_(object, hyperparameter_dictionary)
    return oce.create_BC(mp)


def optimize(
    model: Union[BaseModel, dict],
    runner: Union[BaseModelManager, Callable],
    max_evals=3,
):
    hyperparameter_index = index_hyperparameters(model)

    def objective(hyperparameter_dictionary, model = model, runner = runner):
        print(hyperparameter_dictionary)
        model = load_hyperparameters(model, hyperparameter_dictionary)
        if issubclass(type(runner), BaseModelManager):
            metric = runner.run(model)
        else:
            metric = runner(model)
        return metric

    best = fmin(
        fn=objective,
        space=hyperparameter_index,
        algo=tpe.suggest,
        max_evals=max_evals,
    )
    return best
